@@ -6,7 +6,7 @@ use warnings;
 
 our $VERSION;
 BEGIN {
-    $VERSION = '0.10';
+    $VERSION = '0.11';
     require XSLoader;
     XSLoader::load('Crypt::Sodium', $VERSION);
     
@@ -43,6 +43,10 @@ our @EXPORT = qw(
     crypto_hash
     crypto_generichash
     crypto_generichash_key
+    crypto_generichash_init
+    crypto_generichash_update
+    crypto_generichash_final
+    crypto_generichash_statebytes
     crypto_stream
     crypto_stream_xor
     crypto_box_keypair
@@ -60,6 +64,7 @@ our @EXPORT = qw(
 
     randombytes_buf
     randombytes_random
+    randombytes_uniform
     box_keypair
     sign_keypair
 
@@ -118,7 +123,7 @@ use subs qw/
     crypto_scalarmult_BYTES
 /;
 
-
+use Crypt::Sodium::GenericHash::State;
 
 sub crypto_box_nonce {
     return randombytes_buf(crypto_box_NONCEBYTES);
@@ -202,7 +207,9 @@ sub crypto_generichash {
         die "[fatal]: key must be between " . crypto_generichash_BYTES_MIN . " and " . crypto_generichash_BYTES_MAX . " bytes long.\n";
     }
 
-    return real_crypto_generichash($to_hash, length($to_hash), $outlen, 0, 0);
+    no warnings 'uninitialized';
+
+    return real_crypto_generichash($to_hash, length($to_hash), $outlen, undef, 0);
 }
 
 sub crypto_generichash_key {
@@ -221,6 +228,36 @@ sub crypto_generichash_key {
     return real_crypto_generichash($to_hash, length($to_hash), $outlen, $key, length($key));
 }
 
+sub crypto_generichash_init {
+    my ($key, $hash_size) = @_;
+
+    if ($key) {
+        unless ((length($key) >= crypto_generichash_KEYBYTES_MIN) &&
+            (length($key) <= crypto_generichash_KEYBYTES_MAX)) {
+            die "[fatal]: key must be undef or between " . crypto_generichash_KEYBYTES_MIN . " and " . crypto_generichash_KEYBYTES_MAX . " bytes long.\n";
+        }
+    }
+
+    $hash_size = 64 unless $hash_size;
+
+    no warnings 'uninitialized';
+
+    return Crypt::Sodium::GenericHash::State->new(
+        state => real_crypto_generichash_init($key, length($key), $hash_size),
+        outlen => $hash_size,
+    );
+}
+
+sub crypto_generichash_update {
+    my ($state, $to_append) = @_;
+    real_crypto_generichash_update($state->{state}, $to_append, length($to_append));
+}
+
+sub crypto_generichash_final {
+    my ($state) = @_;
+    return real_crypto_generichash_final($state->{state}, $state->{outlen});
+}
+
 sub box_keypair {
     my $ar = crypto_box_keypair();
     return (@$ar);
